@@ -32,12 +32,24 @@
 #include <cairo/cairo-ft.h>
 #endif /* CAIRO_HAS_FC_FONT */
 
+#ifdef HAVE_GS
+#include <ghostscript/ierrors.h>
+#include <ghostscript/iapi.h>
+#include <ghostscript/gdevdsp.h>
+#endif /* HAVE_GS */
+
 extern XCWindowData *areawin;
-extern int *appcolors;
 extern Globaldata xobjs;
+extern int number_colors;
+extern colorindex *colorlist;
 extern short fontcount;
 extern fontinfo *fonts;
+extern Cursor appcursors[NUM_CURSORS];
 extern Display *dpy;
+extern gs_state_t gs_state;
+extern cairo_surface_t *bbuf;
+
+static cairo_user_data_key_t fontinfo_key;
 
 static void xc_cairo_strokepath(short style, float width);
 
@@ -45,23 +57,27 @@ static void xc_cairo_strokepath(short style, float width);
 
 void xc_cairo_set_matrix(const Matrix *xcm)
 {
-   cairo_matrix_t m = {
-	 .xx = xcm->a, .xy = xcm->b, .x0 = xcm->c,
-	 .yx = xcm->d, .yy = xcm->e, .y0 = xcm->f
-   };
+   cairo_matrix_t m;
+   m.xx = xcm->a; m.xy = xcm->b; m.x0 = xcm->c;
+   m.yx = xcm->d; m.yy = xcm->e; m.y0 = xcm->f;
    cairo_set_matrix(areawin->cr, &m);
-};
+}
 
 /*----------------------------------------------------------------------*/
 /* Set the color, based on the given color index                        */
 /*----------------------------------------------------------------------*/
 
-void xc_cairo_set_color(unsigned long coloridx)
+void xc_cairo_set_color(int coloridx)
 {
-   unsigned short r, g, b;
-   double red, green, blue;
-   xc_get_color_rgb(coloridx, &r, &g, &b);
-   cairo_set_source_rgb(areawin->cr, r / 65535., g / 65535., b / 65535.);
+   colorindex *xcc;
+
+   if (coloridx >= number_colors) return;
+
+   xcc = &colorlist[coloridx];
+   cairo_set_source_rgb(areawin->cr,
+		(double)xcc->color.red / 65535.,
+		(double)xcc->color.green / 65535.,
+		(double)xcc->color.blue / 65535.);
 }
 
 /*----------------------------------------------------------------------------*/
@@ -240,10 +256,9 @@ void UDrawBox(XPoint origin, XPoint corner)
 /* closely reach the position "corner".					*/
 /*----------------------------------------------------------------------*/
 
-float UDrawRescaleBox(XPoint *corner)
+void UDrawRescaleBox(XPoint *corner)
 {
    XPoint newpoints[5];
-   float newscale;
 
    if (!areawin->redraw_ongoing) {
       areawin->redraw_needed = True;
@@ -251,12 +266,12 @@ float UDrawRescaleBox(XPoint *corner)
    }
 
    if (areawin->selects == 0)
-      return 0.;
-
-   newscale = UGetRescaleBox(corner, newpoints);
+      return;
 
    if (areawin->redraw_ongoing) {
       int i;
+      UGetRescaleBox(corner, newpoints);
+
       cairo_save(areawin->cr);
       xc_cairo_set_color(AUXCOLOR);
       cairo_set_dash(areawin->cr, NULL, 0, 0.);
@@ -268,8 +283,6 @@ float UDrawRescaleBox(XPoint *corner)
       xc_cairo_strokepath(0, 1);
       cairo_restore(areawin->cr);
    }
-   
-   return newscale;
 }
 
 /*-------------------------------------------------------------------------*/
@@ -450,8 +463,9 @@ void UDrawObject(objinstptr theinstance, short level, int passcolor,
 
      /* restore the color passed to the object, if different from current color */
 
-     if ((defaultcolor != DOFORALL) && (passcolor != curcolor))
+     if ((defaultcolor != DOFORALL) && (passcolor != curcolor)) {
 	XTopSetForeground(passcolor);
+     }
    }
 
    /* restore the selection list (if any) */
@@ -463,8 +477,6 @@ void UDrawObject(objinstptr theinstance, short level, int passcolor,
 
 static void xc_cairo_strokepath(short style, float width)
 {
-   float tmpwidth;
-
    if (!(style & CLIPMASK) || (areawin->showclipmasks == TRUE)) {
       if (style & FILLED || (!(style & FILLED) && style & OPAQUE)) {
          if ((style & FILLSOLID) == FILLSOLID)
@@ -495,7 +507,8 @@ static void xc_cairo_strokepath(short style, float width)
 	 cairo_set_line_join(areawin->cr, (style & SQUARECAP) ? 
 	       CAIRO_LINE_JOIN_MITER : CAIRO_LINE_JOIN_BEVEL);
          if (style & (DASHED | DOTTED)) {
-            double dashes[2] = {4.0 * width, 4.0 * width};
+            double dashes[2];
+	    dashes[0] = dashes[1] = 4.0 * width;
 	    if (style & DOTTED)
 	       dashes[0] = width;
 	    cairo_set_dash(areawin->cr, dashes, 2, 0.0);
@@ -516,7 +529,7 @@ static void xc_cairo_strokepath(short style, float width)
    }
    if (style & CLIPMASK)
       cairo_clip_preserve(areawin->cr);
-   cairo_new_path(areawin->cr); // clear preserved paths
+   cairo_new_path(areawin->cr); /* clear preserved paths */
 }
 
 /*-------------------------------------------------------------------------*/
@@ -542,22 +555,17 @@ void UDrawPolygon(polyptr thepoly, float passwidth)
 
 void UDrawArc(arcptr thearc, float passwidth)
 {
-   XPoint tmppoints[RSTEPS + 2];
-   float scaledwidth;
-
    if (!areawin->redraw_ongoing) {
       areawin->redraw_needed = True;
       return;
    }
 
-   scaledwidth = thearc->width * passwidth;
-
    if (abs(thearc->radius) == thearc->yaxis)
       cairo_arc(areawin->cr, thearc->position.x, thearc->position.y,
 	    abs(thearc->radius), thearc->angle1 * M_PI / 180.0,
 	    thearc->angle2 * M_PI / 180.0);
-   else {
-      // perform elliptical arc, as described in cairo manual
+   else if (thearc->yaxis) {
+      /* perform elliptical arc, as described in cairo manual */
       cairo_save(areawin->cr);
       cairo_translate(areawin->cr, thearc->position.x, thearc->position.y);
       cairo_scale(areawin->cr, abs(thearc->radius), thearc->yaxis);
@@ -565,10 +573,22 @@ void UDrawArc(arcptr thearc, float passwidth)
 	    thearc->angle2 * M_PI / 180.0);
       cairo_restore(areawin->cr);
    }
-   xc_cairo_strokepath(thearc->style, thearc->width * passwidth);
-   if (thearc->cycle != NULL) {
-      UDrawXLine(thearc->position, areawin->save);
+   else { /* no y-axis dimension: draw a line */
+      /* can we do this in a more elegant manner? */
+      double theta;
+      double theta_start = thearc->angle1 * RADFAC;
+      double theta_stop = thearc->angle2 * RADFAC;
+      cairo_move_to(areawin->cr, thearc->position.x + fabs(thearc->radius)
+	    * cos(theta_start), thearc->position.y);
+      for (theta = -M_PI; theta < theta_stop; theta += M_PI) {
+         if (theta <= theta_start) continue;
+         cairo_line_to(areawin->cr, thearc->position.x 
+	    + fabs(thearc->radius) * cos(theta), thearc->position.y);
+      }
+      cairo_line_to(areawin->cr, thearc->position.x + fabs(thearc->radius)
+	    * cos(theta_stop), thearc->position.y);
    }
+   xc_cairo_strokepath(thearc->style, thearc->width * passwidth);
 }
 
 /*-------------------------------------------------------------------------*/
@@ -578,28 +598,12 @@ void UDrawPath(pathptr thepath, float passwidth)
    genericptr	*genpath;
    polyptr	thepoly;
    splineptr	thespline;
-   Boolean	draweditlines = FALSE;
    
    if (!areawin->redraw_ongoing) {
       areawin->redraw_needed = True;
       return;
    }
 
-   /* First pass---check for any splines that are being edited.  If	*/
-   /* any one is, then draw all control points for all splines in the	*/
-   /* path.								*/
-
-   for (genpath = thepath->plist; genpath < thepath->plist + thepath->parts;
-	  genpath++) {
-      if (ELEMENTTYPE(*genpath) == SPLINE) {
-	 thespline = TOSPLINE(genpath);
-	 if (thespline->cycle != NULL) {
-	    draweditlines = TRUE;
-	    break;
-         }
-      }
-   }
-
    /* Draw first point */
    if (thepath->parts) {
       genpath = thepath->plist;
@@ -637,17 +641,6 @@ void UDrawPath(pathptr thepath, float passwidth)
       }
    }
    xc_cairo_strokepath(thepath->style, thepath->width * passwidth);
-   /* Finally draw the edit lines */ 
-   if (draweditlines) {
-      for (genpath = thepath->plist; genpath < thepath->plist + thepath->parts;
-	    genpath++) {
-         if (ELEMENTTYPE(*genpath) == SPLINE) {
-	    thespline = TOSPLINE(genpath);
-	    UDrawXLine(thespline->ctrl[0], thespline->ctrl[1]);  
-	    UDrawXLine(thespline->ctrl[3], thespline->ctrl[2]);
-         }
-      }
-   }
 }
 
 /*-------------------------------------------------------------------------*/
@@ -664,10 +657,6 @@ void UDrawSpline(splineptr thespline, float passwidth)
          thespline->ctrl[2].x, thespline->ctrl[2].y,
 	 thespline->ctrl[3].x, thespline->ctrl[3].y);
    xc_cairo_strokepath(thespline->style, thespline->width * passwidth);
-   if (thespline->cycle != NULL) {
-      UDrawXLine(thespline->ctrl[0], thespline->ctrl[1]);  
-      UDrawXLine(thespline->ctrl[3], thespline->ctrl[2]);
-   }
 }
 
 /*-------------------------------------------------------------------------*/
@@ -699,11 +688,9 @@ void UDrawGraphic(graphicptr gp)
 void draw_grids(void)
 {
    double spc, spc2, spc3;
-   cairo_matrix_t m = {
-	 .xx = 1., .yx = 0., .xy = 0., .yy = -1., 
-	 .x0 = -areawin->pcorner.x * areawin->vscale,
-	 .y0 = areawin->height + areawin->pcorner.y * areawin->vscale
-   };
+   cairo_matrix_t m = {1., 0., 0., -1., 0., 0.};
+   m.x0 = -areawin->pcorner.x * areawin->vscale;
+   m.y0 = areawin->height + areawin->pcorner.y * areawin->vscale;
 
    if (!areawin->redraw_ongoing) {
       areawin->redraw_needed = True;
@@ -822,115 +809,203 @@ void draw_grids(void)
 }
 
 /*---------------------------------------------------------------------*/
-/* draw a single character at 0, 0 using current transformation matrix */
+/* Draw a string at position offset, starting at the start index and   */
+/* ending before the end index. The offset is updated on return.       */
 /*---------------------------------------------------------------------*/
 
-float UDrawChar(u_char code, short styles, short ffont, int groupheight,
-	int passcolor, float passwidth)
+void UDrawCharString(u_char *text, int start, int end, XfPoint *offset,
+      short styles, short ffont, int groupheight, int passcolor, float tmpscale)
 {
-   objectptr drawchar;
-   XPoint alphapts[2];
-   float localwidth;
-   objinst charinst;
+   int idx, nglyphs = 0;
+   cairo_matrix_t oldm;
+   double x, y;
+   cairo_glyph_t *glyphs;
+   UNUSED(passcolor);
 
-   if (!areawin->redraw_ongoing) {
-      areawin->redraw_needed = True;
-      /* No return here, since the return value might be needed? */
-   }
-
-   if ((ffont >= fontcount) || (fonts[ffont].encoding == NULL))
-      return 0;
-
-   alphapts[0].x = 0;
-   alphapts[0].y = 0;
-   charinst.type = OBJINST;
-   charinst.color = DEFAULTCOLOR;
-   charinst.rotation = 0;
-   charinst.scale = fonts[ffont].scale;
-   charinst.position = alphapts[0];
-   charinst.params = NULL;
-   
-   /* get proper font and character */
-
-   drawchar = fonts[ffont].encoding[(u_char)code];
-   charinst.thisobject = drawchar;
-
-   if (fonts[ffont].font_face) {
-      cairo_save(areawin->cr);
-      cairo_text_extents_t extents;
-      cairo_set_font_face(areawin->cr, fonts[ffont].font_face);
-   }
-
-   if (!fonts[ffont].font_face && (fonts[ffont].flags & 0x22) == 0x22) {
-      /* font is not a cairo font and it is derived and italic */
-      USlantCTM(DCTM, 0.25);  		/* premultiply by slanting function */
-   }
-
-   /* Draw glyph */
-   if (!(styles & 64) && areawin->redraw_ongoing) {
-      if (fonts[ffont].font_face) {
-	 cairo_set_font_size (areawin->cr, 40.); /* TODO: Why 40? */
-	 cairo_scale(areawin->cr, 1., -1.);
-	 cairo_move_to (areawin->cr, 0., 0.);
-	 cairo_show_text(areawin->cr, fonts[ffont].utf8encoding[code]);
-   	 cairo_new_path(areawin->cr);
-      }	
-      else
-	 UDrawObject(&charinst, SINGLE, passcolor, passwidth, NULL);
-   }
-
-   /* Determine character width */
-   if (fonts[ffont].font_face) {
-      /* Determine localwidth on a fixed font size of 100, to prevent hinting */
-      /* from destroying scale independance */
-      cairo_text_extents_t extents;
-      cairo_identity_matrix(areawin->cr);
-      cairo_set_font_size(areawin->cr, 100.);
-      cairo_text_extents(areawin->cr, fonts[ffont].utf8encoding[code],
-	    &extents);
-      localwidth = extents.x_advance * fonts[ffont].scale / 100. * 40.; /* TODO: Why 40? */
-      cairo_restore(areawin->cr);
-   }
-   else
-      localwidth = (drawchar->bbox.lowerleft.x + drawchar->bbox.width)
-	    * fonts[ffont].scale;
+   cairo_get_matrix(areawin->cr, &oldm);
+   cairo_scale(areawin->cr, tmpscale, fabs(tmpscale));
 
    /* under- and overlines */
-   if (!(styles & 64)) {
-      if (styles & 8)
-         alphapts[0].y = alphapts[1].y = -6;
-      else if (styles & 16)
-         alphapts[0].y = alphapts[1].y = groupheight + 4;
-      if (styles & 24 && areawin->redraw_ongoing) {
-         alphapts[0].x = 0; alphapts[1].x = localwidth;
-	 cairo_set_line_width(areawin->cr, passwidth);
-	 cairo_move_to(areawin->cr, alphapts[0].x, alphapts[0].y);
-	 cairo_line_to(areawin->cr, alphapts[1].x, alphapts[1].y);
-	 cairo_stroke(areawin->cr);
-      }
+   if (styles & 8)
+      y = offset->y / tmpscale - 6.;
+   else if (styles & 16)
+      y = offset->y / tmpscale + groupheight + 4;
+   if (styles & 24) {
+      /* under/overline thickness scales if bold */
+      float tmpthick = ((fonts[ffont].flags & 0x21) == 0x21) ?  4.0 : 2.0;
+      cairo_set_line_width(areawin->cr, tmpthick);
+
+      x = offset->x / tmpscale;
    }
-   return localwidth;
+
+   glyphs = cairo_glyph_allocate(end - start);
+   for (idx = start; idx < end; idx++)
+   {
+      /* Add glyphs */
+      glyphs[nglyphs].index = fonts[ffont].glyph_index[text[idx]];
+      glyphs[nglyphs].x = offset->x / tmpscale;
+      glyphs[nglyphs].y = offset->y / fabs(tmpscale);
+      nglyphs++;
+      /* Determine character width */
+      offset->x += fonts[ffont].glyph_advance[text[idx]] * tmpscale;
+   }
+   if (nglyphs) {
+      cairo_show_glyphs(areawin->cr, glyphs, nglyphs);
+      cairo_new_path(areawin->cr);
+   }
+   cairo_glyph_free(glyphs);
+   
+   if (styles & 24) {
+      cairo_move_to(areawin->cr, x, y);
+      cairo_line_to(areawin->cr, offset->x / tmpscale, y);
+      cairo_stroke(areawin->cr);
+   }
+   cairo_set_matrix(areawin->cr, &oldm);
 }
 
-float xc_cairo_get_char_extents(const fontinfo *font, unsigned char c,
-      float *top, float *bottom)
-{
-   /* Determine localwidth on a fixed font size of 100, to */
-   /* prevent hinting from destroying scale independance */
-   cairo_text_extents_t extents;
-   cairo_save(areawin->cr);
-   cairo_identity_matrix(areawin->cr);
-   cairo_set_font_face(areawin->cr, font->font_face);
-   cairo_set_font_size(areawin->cr, 100.);
-   cairo_text_extents(areawin->cr, font->utf8encoding[c], &extents);
-   cairo_restore(areawin->cr);
+/*----------------------------------------------------------------------*/
+/* A completely stripped down version of UDrawObject, used for stroke	*/
+/* font drawing in xc_user_font_render					*/
+/*----------------------------------------------------------------------*/
 
-   if (top)
-      *top = -extents.y_bearing * 40. / 100.;
-   if (bottom)
-      *bottom = (extents.height - extents.y_bearing) * 40. / 100.;
-   return extents.x_advance * 40. / 100.; /* TODO: Why 40? */
-}	
+static void xc_draw_glyph_object(objinstptr inst, float passwidth)
+{
+   int thispart;
+   genericptr *areagen;
+   objectptr obj = inst->thisobject;
+   objinstptr ptr;
+
+   for (thispart = 0; thispart < obj->parts; thispart++) {
+      areagen = obj->plist + thispart;
+      switch(ELEMENTTYPE(*areagen)) {
+	 case(POLYGON): 
+            UDrawPolygon(TOPOLY(areagen), passwidth);
+	    break;
+	 case(SPLINE):
+            UDrawSpline(TOSPLINE(areagen), passwidth);
+	    break;
+	 case(ARC):
+            UDrawArc(TOARC(areagen), passwidth);
+	    break;
+	 case(PATH):
+	    UDrawPath(TOPATH(areagen), passwidth);
+	    break;
+         case(OBJINST):
+	    ptr = TOOBJINST(areagen);
+	    cairo_save(areawin->cr);
+	    cairo_translate(areawin->cr, ptr->position.x, ptr->position.y);
+	    cairo_rotate(areawin->cr, -ptr->rotation * RADFAC);
+	    cairo_scale(areawin->cr, ptr->scale, fabs(ptr->scale));
+	    xc_draw_glyph_object(ptr, passwidth);
+	    cairo_restore(areawin->cr);
+	    break;
+      }
+   }
+}
+
+/*----------------------------------------------------------------------*/
+/* Rendering function for a cairo_user_font_face, to draw the stroke	*/
+/* fonts of xcircuit							*/
+/* TODO: always this factor 40 to match with cairo. Is it BASELINE?	*/
+/*----------------------------------------------------------------------*/
+
+static cairo_status_t xc_user_font_render(cairo_scaled_font_t *scaled_font,
+      unsigned long glyph, cairo_t *cr, cairo_text_extents_t *extents)
+{
+   objectptr chr;
+   objinst charinst;
+   objinstptr theinstance;
+   cairo_t *old_cr;
+   float passwidth;
+   double llx, lly, trx, try;
+
+   cairo_font_face_t *ff = cairo_scaled_font_get_font_face(scaled_font);
+   size_t fontidx = (size_t) cairo_font_face_get_user_data(ff, &fontinfo_key);
+   fontinfo *fi = &fonts[fontidx];
+	
+   chr = fi->encoding[glyph];
+   theinstance = &charinst;
+   charinst.thisobject = chr;
+
+   llx = chr->bbox.lowerleft.x / 40.;
+   lly = chr->bbox.lowerleft.y / 40.;
+   trx = (chr->bbox.lowerleft.x + chr->bbox.width) / 40.;
+   try = (chr->bbox.lowerleft.y + chr->bbox.height) / 40.;
+
+   /* temporary override areawin->cr with the font context */
+   old_cr = areawin->cr;
+   areawin->cr = cr;
+
+   cairo_scale(cr, 1. / 40., -1. / 40.);
+   cairo_set_line_width(cr, 1);
+
+   /* if font is derived and italic, premultiply by slanting matrix */
+   if ((fi->flags & 0x22) == 0x22) {
+      cairo_matrix_t m = {1., 0., .25, 1., 0., 0.};
+      cairo_transform(areawin->cr, &m);
+      llx += .25 * lly;
+      trx += .25 * try;
+   }
+   /* simple boldface technique for derived fonts */
+   passwidth = ((fi->flags & 0x21) == 0x21) ?  4. : 2.;
+
+   /* Correct extentswith line width */
+   llx -= passwidth / 40.;
+   lly -= passwidth / 40.;
+   trx += passwidth / 40.;
+   try += passwidth / 40.;
+
+   xc_draw_glyph_object(theinstance, passwidth);
+
+   extents->x_bearing = llx;
+   extents->y_bearing = -try;
+   extents->width = trx - llx;
+   extents->height = try - lly;
+   extents->x_advance = (chr->bbox.lowerleft.x + chr->bbox.width) / 40.;
+   extents->y_advance = 0.;
+   
+   areawin->cr = old_cr;
+   return CAIRO_STATUS_SUCCESS;
+}
+
+/*----------------------------------------------------------------------*/
+/* Function to translate unicode into a glyph index for the stroke font */
+/*----------------------------------------------------------------------*/
+
+static cairo_status_t xc_user_font_glyph(cairo_scaled_font_t *scaled_font,
+      unsigned long unicode, unsigned long *glyph_index)
+{
+   cairo_font_face_t *ff = cairo_scaled_font_get_font_face(scaled_font);
+   size_t fontidx = (size_t) cairo_font_face_get_user_data(ff, &fontinfo_key);
+   fontinfo *fi = &fonts[fontidx];
+   unsigned long idx; 
+
+   /* search all glyphs in the utf8encoding. This is allowed to be slow, */
+   /* as the glyph list will be buffered anyway */
+   for (idx = 1; idx < 255; idx++) {
+      const char *s = fi->utf8encoding[idx];
+      int cidx = 0;
+
+      /* Convert utf-8 to unicode */
+      unsigned long uc = s[0];
+      if (uc & 0x80) {
+	 unsigned long mask = 0x3f;
+	 while ((s[++cidx] & 0xc0) == 0x80) {
+	    uc = (uc << 6) | (s[cidx] & 0x3f);
+	    mask = (mask << 5) | 0x1f;
+	 }
+	 uc &= mask;
+      }
+
+      if (unicode == uc) {
+	 *glyph_index = idx;
+	 return CAIRO_STATUS_SUCCESS;
+      }
+   }
+   /* This should not happen: replace unknown glyph with question mark */
+   *glyph_index = '?';
+   return CAIRO_STATUS_SUCCESS;
+}
 
 /*---------------------------------------------------------------------*/
 /* find the corresponing cairo fontface for a given fontinfo structure */
@@ -939,48 +1014,56 @@ float xc_cairo_get_char_extents(const fontinfo *font, unsigned char c,
 typedef struct {
    const char* postscript_name;
    const char* replacement_name;
+   const char* foundry_name;
 } xc_font_replacement;
 
-static const xc_font_replacement urw_fonts[] =
+static const xc_font_replacement replacement_fonts[] =
 {
-   {"ITC Avant Garde Gothic",	"URW Gothic L"},
-   {"ITC Bookman",		"URW Bookman L"},
-   {"Courier",			"Nimbus Mono L"},
-   {"Helvetica",		"Nimbus Sans L"},
-   {"Helvetica Narrow",		"Nimbus Sans L Condensed"},
-   {"New Century Schoolbook",	"Century Schoolbook L"},
-   {"Palatino",			"URW Palladio L"},
-   {"Symbol",			"Standard Symbols L"},
-   {"Times",			"Nimbus Roman No9 L"},
-   {"Times-Roman",		"Nimbus Roman No9 L"},
-   {"ITC ZapfChangery",		"URW Changery L"},
-   {"ITC ZapfDingbats",		"Dingbats"},
-   {NULL, NULL}
+   /* First try to see if 'real' postscript fonts have been installed */
+   {"ITC Avant Garde Gothic",	"ITC Avant Garde Gothic",   "adobe"},
+   {"ITC Bookman",		"ITC Bookman",		    "adobe"},
+   {"Courier",			"Courier",		    "adobe"},
+   {"Helvetica",		"Helvetica",		    "adobe"},
+   {"Helvetica Narrow",		"Helvetica Narrow",	    "adobe"},
+   {"New Century Schoolbook",	"New Century Schoolbook",   "adobe"},
+   {"Palatino",			"Palatino",		    "adobe"},
+   {"Symbol",			"Symbol",		    "adobe"},
+   {"Times",			"Times",		    "adobe"},
+   {"Times-Roman",		"Times-Roman",		    "adobe"},
+   {"ITC ZapfChangery",		"ITC ZapfChangery",	    "adobe"},
+   {"ITC ZapfDingbats",		"ITC ZapfDingbats",	    "adobe"},
+   /* Next try the URW postscript fonts (guaranteed to have same extents) */
+   {"ITC Avant Garde Gothic",	"URW Gothic L",		    "urw"},
+   {"ITC Bookman",		"URW Bookman L",	    "urw"},
+   {"Courier",			"Nimbus Mono L",	    "urw"},
+   {"Helvetica",		"Nimbus Sans L",	    "urw"},
+   {"Helvetica Narrow",		"Nimbus Sans L Condensed",  "urw"},
+   {"New Century Schoolbook",	"Century Schoolbook L",	    "urw"},
+   {"Palatino",			"URW Palladio L",	    "urw"},
+   {"Symbol",			"Standard Symbols L",	    "urw"},
+   {"Times",			"Nimbus Roman No9 L",	    "urw"},
+   {"Times-Roman",		"Nimbus Roman No9 L",	    "urw"},
+   {"ITC ZapfChangery",		"URW Changery L",	    "urw"},
+   {"ITC ZapfDingbats",		"Dingbats",		    "urw"},
+   /* No success, use the 'old' stroke fonts */
+   {NULL, NULL, NULL}
 };
 
-static const xc_font_replacement css_fonts[] =
-{
-   {"Courier",			"monospace"},
-   {"Helvetica",		"sans-serif"},
-   {"Times",			"serif"},
-   {"Times-Roman",		"serif"},
-   {NULL, NULL}
-};
-
-void xc_cairo_set_fontinfo(fontinfo *fi)
+void xc_cairo_set_fontinfo(size_t fontidx)
 {
    /* TODO: memory leak. font_face is created here. It should also be */
    /* destroyed again somewhere */
+   fontinfo *fi = &fonts[fontidx];
    const char *family = fi->family;
    const xc_font_replacement *replace;
+   int c;
+   
 #ifdef CAIRO_HAS_FC_FONT
    int weight = FC_WEIGHT_NORMAL;
    int slant = FC_SLANT_ROMAN;
-   FcPattern *pattern;
-   FcPattern *matched;
-   FcResult result;
-   FcChar8 *matched_family;
 
+   fi->font_face = NULL;
+   
    if (fi->flags & 1)
       weight = FC_WEIGHT_BOLD;
 
@@ -990,69 +1073,79 @@ void xc_cairo_set_fontinfo(fontinfo *fi)
       else
          slant = FC_SLANT_ITALIC;
    }
-   /* Try the URW fonts first to ensure proper postscript font dimensions */
-   for (replace = urw_fonts; replace->postscript_name; replace++) {
+   /* Try to find a proper postscript font */
+   for (replace = replacement_fonts; replace->postscript_name; replace++) {
       if (!strcmp(replace->postscript_name, fi->family)) {
-	 family = replace->replacement_name;
-	 break;
-      }
-   }
-   pattern = FcPatternBuild (NULL,
-         FC_FAMILY, FcTypeString, family,
-	 FC_WEIGHT, FcTypeInteger, weight,
-	 FC_SLANT, FcTypeInteger, slant,
-	 FC_FOUNDRY, FcTypeString, "urw",
-	 NULL);
-   FcConfigSubstitute(0, pattern, FcMatchPattern);
-   FcDefaultSubstitute(pattern);
-   matched = FcFontMatch(0, pattern, &result);
-
-   /* Check if the matched font is actually the replacement font */
-   FcPatternGetString(matched, FC_FAMILY, 0, &matched_family);
-   if (strcmp(matched_family, family)) {
-      /* No exact match, retry with css font family names */
-      for (replace = css_fonts; replace->postscript_name; replace++) {
-         if (!strcmp(replace->postscript_name, fi->family)) {
-	    family = replace->replacement_name;
+	 FcPattern *matched;
+	 FcChar8 *matched_family, *matched_foundry;
+	 FcResult result;
+	 FcPattern *pattern = FcPatternBuild(NULL,
+	       FC_FAMILY, FcTypeString, replace->replacement_name,
+	       FC_WEIGHT, FcTypeInteger, weight,
+	       FC_SLANT, FcTypeInteger, slant,
+	       FC_FOUNDRY, FcTypeString, replace->foundry_name,
+	       NULL);
+	 FcConfigSubstitute(0, pattern, FcMatchPattern);
+	 FcDefaultSubstitute(pattern);
+	 matched = FcFontMatch(0, pattern, &result);
+	 /* Check if the matched font is actually the replacement font */
+	 FcPatternGetString(matched, FC_FAMILY, 0, &matched_family);
+	 FcPatternGetString(matched, FC_FOUNDRY, 0, &matched_foundry);
+	 if (!strcmp(matched_family, replace->replacement_name)
+	       && !strcmp(matched_foundry, replace->foundry_name))
+	    fi->font_face = cairo_ft_font_face_create_for_pattern(matched);
+	 FcPatternDestroy(matched);
+	 FcPatternDestroy(pattern);
+	 if (fi->font_face)
 	    break;
-         }
       }
-      pattern = FcPatternBuild (NULL,
-            FC_FAMILY, FcTypeString, family,
-	    FC_WEIGHT, FcTypeInteger, weight,
-	    FC_SLANT, FcTypeInteger, slant,
-	    FC_OUTLINE, FcTypeBool, FcTrue,
-	    NULL);
-      FcConfigSubstitute(0, pattern, FcMatchPattern);
-      FcDefaultSubstitute(pattern);
-      matched = FcFontMatch(0, pattern, &result);
    }
-   fi->font_face = cairo_ft_font_face_create_for_pattern(matched);
-
-   FcPatternDestroy(matched);
-   FcPatternDestroy(pattern);
 #else /* CAIRO_HAS_FC_FONT */
-   int weight = CAIRO_FONT_WEIGHT_NORMAL;
-   int slant = CAIRO_FONT_SLANT_NORMAL;
+   fi->font_face = NULL;
+#endif /* CAIRO_HAS_FC_FONT */
 
-   if (fi->flags & 1)
-      weight = CAIRO_FONT_WEIGHT_BOLD;
+   /* Cache the dimensions and all glyphs of the font */
+   if (fi->font_face) {
+      int num_glyphs;
+      cairo_glyph_t *glyphs = NULL;
+      cairo_scaled_font_t *scaled_font;
+      cairo_text_extents_t extents;
 
-   if (fi->flags & 2) {
-      if (!strcmp(family, "Helvetica"))
-         slant = CAIRO_FONT_SLANT_OBLIQUE;
-      else
-         slant = CAIRO_FONT_SLANT_ITALIC;
+      cairo_save(areawin->cr);
+      cairo_identity_matrix(areawin->cr);
+      cairo_set_font_face(areawin->cr, fi->font_face);
+      cairo_set_font_size(areawin->cr, 100.);
+      scaled_font = cairo_get_scaled_font(areawin->cr);
+      for (c = 1; c < 256; c++) { /* skip encoding at index 0 */
+	 cairo_scaled_font_text_to_glyphs(scaled_font, 0., 0.,
+	       fi->utf8encoding[c], -1, &glyphs, &num_glyphs,
+	       NULL, NULL, NULL);
+	 fi->glyph_index[c] = glyphs[0].index;
+	 cairo_scaled_font_glyph_extents(scaled_font, glyphs, 1, &extents);
+	 fi->glyph_top[c] = -extents.y_bearing * 40. / 100.;
+	 fi->glyph_bottom[c] = extents.height * 40. / 100. + fi->glyph_top[c];
+	 fi->glyph_advance[c] = extents.x_advance * 40. / 100.;
+      }
+      cairo_glyph_free(glyphs);
+      cairo_restore(areawin->cr);
    }
-   /* Use the css font family names for the toy font api */
-   for (replace = css_fonts; replace->postscript_name; replace++) {
-      if (!strcmp(replace->postscript_name, fi->family)) {
-	 family = replace->replacement_name;
-	 break;
+   else {
+      /* No decent postscript font found. Backup using stroke fonts */
+      fi->font_face = cairo_user_font_face_create();
+      cairo_font_face_set_user_data(fi->font_face, &fontinfo_key,
+	    (void*) fontidx, (cairo_destroy_func_t) cairo_font_face_destroy);
+      cairo_user_font_face_set_render_glyph_func(fi->font_face,
+	    xc_user_font_render);
+      cairo_user_font_face_set_unicode_to_glyph_func(fi->font_face,
+	    xc_user_font_glyph);
+      for (c = 0; c < 256; c++) {
+	 objectptr chr = fi->encoding[c];
+	 fi->glyph_index[c] = c;
+	 fi->glyph_top[c] = chr->bbox.lowerleft.y + chr->bbox.height;
+	 fi->glyph_bottom[c] = chr->bbox.lowerleft.y;
+	 fi->glyph_advance[c] = chr->bbox.lowerleft.x + chr->bbox.width;
       }
    }
-   fi->font_face = cairo_toy_font_face_create(family, slant, weight);
-#endif /* CAIRO_HAS_FC_FONT */
 }
 
 /*----------------------------------------------------------------------*/
@@ -1106,11 +1199,205 @@ void xcImagePutPixel(xcImage *img, int x, int y, u_char r, u_char g, u_char b)
 void xcImageGetPixel(xcImage *img, int x, int y, u_char *r, u_char *g,
       u_char *b)
 {
+   uint32_t argb;
    xcImageCheckCache(img);
-   uint32_t argb = xcImagePixel_data[y * xcImagePixel_width + x];
+   argb = xcImagePixel_data[y * xcImagePixel_width + x];
    *r = argb >> 16;
    *g = argb >> 8;
    *b = argb;
 }
 
+
+/*------------------------------------------------------------------------*/
+/* Ghostscript rendering function. In contrast to X11, direct calls to    */
+/* ghostscriptapi are made						  */
+/*------------------------------------------------------------------------*/
+
+/*------------------------------------------------------------------------*/
+/* gsapi I/O redirection functions                                        */
+/*------------------------------------------------------------------------*/
+
+#ifdef HAVE_GS
+static int GSDLLCALL gs_stdin_fn(void *caller_handle, char *buf, int len)
+{
+   UNUSED(caller_handle); UNUSED(buf); UNUSED(len);
+
+   return 0; /* EOF */
+}
+
+static int GSDLLCALL gs_stdout_fn(void *caller_handle, const char *str, int len)
+{
+#ifndef GS_DEBUG
+   UNUSED(caller_handle); UNUSED(str);
+#else
+   fwrite(str, 1, len, stdout);
+   fflush(stdout);
+#endif
+   return len;
+}
+
+static int GSDLLCALL gs_stderr_fn(void *caller_handle, const char *str, int len)
+{
+#ifndef GS_DEBUG
+   UNUSED(caller_handle); UNUSED(str);
+#else
+   fwrite(str, 1, len, stderr);
+   fflush(stderr);
+#endif
+   return len;
+}
+#endif /* HAVE_GS */
+
+/*------------------------------------------------------------------------*/
+/* gsapi displaycallback functions                                        */
+/*------------------------------------------------------------------------*/
+
+#ifdef HAVE_GS
+static unsigned char *gs_pimage;
+static int gs_width, gs_height, gs_raster;
+
+static int gs_display_dummy(void)
+{
+   return 0;
+}
+#endif /* HAVE_GS */
+
+#ifdef HAVE_GS
+static int gs_display_size(void *handle, void *device, int width, int height, 
+      int raster, unsigned int format, unsigned char *pimage)
+{
+   UNUSED(handle); UNUSED(device); UNUSED(format);
+
+   gs_pimage = pimage;
+   gs_width = width;
+   gs_height = height;
+   gs_raster = raster;
+   return 0;
+}
+#endif /* HAVE_GS */
+
+#ifdef HAVE_GS
+int gs_display_page(void *handle, void *device, int copies, int flush)
+{
+   cairo_surface_t *tbuf;
+   cairo_t *bbuf_cr;
+   UNUSED(handle); UNUSED(device); UNUSED(copies); UNUSED(flush);
+
+   if (bbuf)
+      cairo_surface_destroy(bbuf);
+   /* Since cairo_image_surface_create_for_data assumes the buffer to be */
+   /* valid until destruction, and ghostscript frees it immediately,     */
+   /* first use a temporary buffer. Then immediately copy the buffer	 */
+   /* To the final location						 */
+   tbuf = cairo_image_surface_create_for_data(gs_pimage, CAIRO_FORMAT_RGB24,
+         gs_width, gs_height, gs_raster);
+   bbuf = cairo_image_surface_create(CAIRO_FORMAT_RGB24, gs_width, gs_height);
+   bbuf_cr = cairo_create(bbuf);
+   cairo_set_source_surface(bbuf_cr, tbuf, 0., 0.);
+   cairo_paint(bbuf_cr);
+   cairo_destroy(bbuf_cr);
+   cairo_surface_destroy(tbuf);
+   return 0;
+}
+#endif /* HAVE_GS */
+
+#ifdef HAVE_GS
+display_callback gs_display = {
+    sizeof(display_callback),
+    DISPLAY_VERSION_MAJOR,
+    DISPLAY_VERSION_MINOR,
+    (int (*)(void*, void*)) gs_display_dummy, /* display_open */
+    (int (*)(void*, void*)) gs_display_dummy, /* display_preclose */
+    (int (*)(void*, void*)) gs_display_dummy, /* display_close */
+    (int (*)(void*, void*, int, int, int, unsigned int)) gs_display_dummy,
+					      /* display_presize */
+    gs_display_size,
+    (int (*)(void*, void*)) gs_display_dummy, /* display_sync */
+    gs_display_page,
+    NULL, /* display_update */
+    NULL, /* display_memalloc */
+    NULL, /* display_memfree */
+    NULL /* display_separation */
+};
+#endif /* HAVE_GS */
+
+/*------------------------------------------------------*/
+/* write scale and position to ghostscript              */
+/* and tell ghostscript to run the requested file       */
+/*------------------------------------------------------*/
+
+#ifdef HAVE_GS
+const char *gs_argv[] = {
+      "-dQUIET",			/* Suppress startup messages	*/
+      "-dNOPAUSE",			/* Disable pause at end of page	*/
+      "-dBATCH",			/* Exit when all files done	*/ 
+      "-sDEVICE=display",
+      "-sDisplayHandle=0",
+      "-r75",				/* Display resolution 		*/
+      "-dGraphicsAlphaBits=4",		/* Graphics anti-aliasing	*/
+      "-dTextAlphaBits=4"		/* Text anti-aliasing		*/
+};
+#endif /* HAVE_GS */
+
+void write_scale_position_and_run_gs(float norm, float xpos, float ypos,
+      const char *bgfile)
+{
+#ifndef HAVE_GS
+   UNUSED(norm); UNUSED(xpos); UNUSED(ypos); UNUSED(bgfile);
+#endif
+#ifdef HAVE_GS
+   int i, code, exit_code;
+   void *instance;
+   char gs_cmd[256];
+   char display_format[] = "-dDisplayFormat=........";
+   char pixmap_size[] = "-g........x........";
+   int argc = sizeof(gs_argv) / sizeof(gs_argv[0]);
+   const char **argv = (const char**) malloc((argc + 2) * sizeof(const char*));
+
+   for (i = 0; i < argc; i++)
+      argv[i] = gs_argv[i];
+   argv[argc++] = display_format;
+   argv[argc++] = pixmap_size;
+
+   sprintf(gs_cmd,
+	 " /GSobj save def"
+   	 " /setpagedevice {pop} def"
+   	 " gsave"
+   	 " %3.2f %3.2f translate"
+   	 " %3.2f %3.2f scale"
+   	 " (%s) run"
+   	 " GSobj restore"
+	 " grestore",
+	 xpos, ypos, norm, norm, bgfile);
+
+   sprintf(display_format, "-dDisplayFormat=%d", DISPLAY_COLORS_RGB
+          | DISPLAY_UNUSED_LAST | DISPLAY_DEPTH_8 | DISPLAY_LITTLEENDIAN
+          | DISPLAY_TOPFIRST | DISPLAY_ROW_ALIGN_DEFAULT);
+   sprintf(pixmap_size, "-g%dx%d", areawin->width, areawin->height);
+
+   XDefineCursor(dpy, areawin->window, WAITFOR);
+
+   if ((code = gsapi_new_instance(&instance, NULL)) == 0) {
+      gsapi_set_stdio(instance, gs_stdin_fn, gs_stdout_fn, gs_stderr_fn);
+      gsapi_set_display_callback(instance, &gs_display);
+      if (!(code = gsapi_init_with_args(instance, argc, (char**) argv)))
+         gsapi_run_string(instance, gs_cmd, 0, &exit_code);
+      gsapi_exit(instance);
+      gsapi_delete_instance(instance);
+   }
+   free(argv);
+#ifdef GS_DEBUG
+   fprintf(stdout, "Xcircuit: ghostscript done\n");
+#endif
+
+   XDefineCursor(dpy, areawin->window, DEFAULTCURSOR);
+   areawin->lastbackground = xobjs.pagelist[areawin->page]->background.name;
+   drawarea(areawin->area, NULL, NULL);
+#endif /* HAVE_GS */
+   gs_state = GS_READY;
+}
+
+#else /* HAVE_CAIRO */
+typedef int no_empty_translation_unit_warning;
 #endif /* HAVE_CAIRO */
+
