@@ -56,8 +56,8 @@ extern Display    *dpy;
 extern GC	  hgc;
 extern XCWindowData *areawin;
 extern ApplicationData appdata;
-extern int 	  *appcolors;
 extern xcWidget     top;
+extern colorindex *colorlist;
 extern short	  popups;
 extern xcWidget	  menuwidgets[];
 extern char *function_names[NUM_FUNCTIONS];
@@ -195,8 +195,8 @@ void printhelppix()
     /* Set up the GC for drawing to the help window pixmap */
 
     if (hgc == NULL) {
-       values.foreground = FOREGROUND;
-       values.background = BACKGROUND;
+       values.foreground = colorlist[FOREGROUND].color.pixel;
+       values.background = colorlist[BACKGROUND].color.pixel;
        values.graphics_exposures = False;
        values.font = appdata.helpfont->fid;
        hgc = XCreateGC(dpy, hwin, GCForeground | GCBackground | GCFont
@@ -260,10 +260,10 @@ void printhelppix()
     helppix = XCreatePixmap(dpy, hwin, helpwidth, helpheight,
 	   	  DefaultDepthOfScreen(DefaultScreenOfDisplay(dpy)));
 
-    XSetForeground(dpy, hgc, FOREGROUND);
+    XSetForeground(dpy, hgc, colorlist[FOREGROUND].color.pixel);
     XFillRectangle(dpy, helppix, hgc, 0, 0, helpwidth, helpheight);
 
-    XSetForeground(dpy, hgc, BACKGROUND);
+    XSetForeground(dpy, hgc, colorlist[BACKGROUND].color.pixel);
     XDrawString(dpy, helppix, hgc, (helpwidth - XTextWidth(appdata.helpfont,
 	helptitle, strlen(helptitle))) >> 1, t1 + 2, helptitle, strlen(helptitle));
     vtmp = lineheight + 15;
@@ -290,7 +290,7 @@ void printhelppix()
        }
        free(bindings);
     }
-    XSetForeground(dpy, hgc, AUXCOLOR);
+    XSetForeground(dpy, hgc, colorlist[AUXCOLOR].color.pixel);
     XDrawLine(dpy, helppix, hgc, 0, t1 + 7, helpwidth, t1 + 7);
 }
 
@@ -458,7 +458,7 @@ void showhsb(xcWidget hsb, caddr_t clientdata, caddr_t calldata)
    XClearArea(dpy, hwin, 0, pstart + pheight, SBARSIZE,
 		hheight - (pstart + pheight), False);
 
-   XSetForeground(dpy, hgc, BARCOLOR);
+   XSetForeground(dpy, hgc, colorlist[BARCOLOR].color.pixel);
    XFillRectangle(dpy, hwin, hgc, 0, pstart, SBARSIZE, pheight);
 }
 
