@@ -48,22 +48,6 @@ extern char *translateencoding(int);
 
 extern char *evaluate_expr(objectptr, oparamptr, objinstptr);
 
-/* from opengl.c:  compatibility functions for OpenGL/X11 */
-
-#ifdef OPENGL
-extern void SetForeground(Display *, GC, int);
-extern void SetBackground(Display *, GC, int);
-extern void SetFunction(Display *, GC, int);
-extern void SetLineAttributes(Display *, GC, float, int, int, int);
-extern void SetDashes(Display *, GC, int, char dashlist[], int);
-extern void SetStipple(Display *, GC, int);
-extern void SetFillStyle(Display *, GC, int);
-extern void DrawLine(Display *, Window, GC, int, int, int, int);
-extern void DrawLines(Display *, Window, GC, XPoint *, int, int);
-extern void DrawPoint(Display *, Window, GC, int, int);
-extern void FillPolygon(Display *, Window, GC, XPoint *, int, int, int);
-#endif
-
 /* from elements.c: */
 
 /* element constructor functions */
@@ -121,6 +105,16 @@ extern void startwire(XPoint *);
 extern void setendpoint(short *, short, XPoint **, XPoint *);
 extern void wire_op(int, int ,int);
 
+void arc_mode_draw(xcDrawType type, arc *newarc);
+void spline_mode_draw(xcDrawType type, spline *newspline);
+void poly_mode_draw(xcDrawType type, polygon *newpoly);
+void path_mode_draw(xcDrawType type, path *newpoly);
+void text_mode_draw(xcDrawType type, label *newlabel);
+void selarea_mode_draw(xcDrawType type, void *unused);
+void rescale_mode_draw(xcDrawType type, void *unused);
+void move_mode_draw(xcDrawType type, void *unused);
+void normal_mode_draw(xcDrawType type, void *unused);
+
 /* from events.c: */
 
 extern Boolean recursefind(objectptr, objectptr);
@@ -233,6 +227,8 @@ extern void finish_op(int, int, int);
 extern void path_op(genericptr, int, int, int);
 extern void inst_op(genericptr, int, int, int);
 extern void resizearea(xcWidget, caddr_t, caddr_t);
+void draw_grids(void);
+void draw_fixed(void);
 extern void drawarea(xcWidget, caddr_t, caddr_t);
 extern void standard_element_delete(short);
 extern void delete_for_xfer(short, short *, int);
@@ -445,7 +441,7 @@ extern void UDrawTLine(labelptr);
 extern void UDrawXLine(XPoint, XPoint);
 extern void UDrawBox(XPoint, XPoint);
 extern float UGetRescaleBox(XPoint *corner, XPoint *newpoints);
-extern float UDrawRescaleBox(XPoint *);
+extern void UDrawRescaleBox(XPoint *);
 extern void UDrawBBox(void);
 extern void strokepath(XPoint *, short, short, float);
 extern void makesplinepath(splineptr, XPoint *);
@@ -858,6 +854,8 @@ extern int renderbackground(void);
 extern int copybackground(void);
 extern int exit_gs(void);
 extern int reset_gs(void);
+void write_scale_position_and_run_gs(float norm, float xpos, float ypos,
+      const char *bgfile);
 
 #ifndef TCL_WRAPPER
 extern Boolean render_client(XEvent *);
@@ -972,7 +970,8 @@ extern Boolean pin_is_bus(labelptr, objinstptr);
 extern int find_cardinal(int, labelptr, objinstptr);
 extern int find_ordinal(int, labelptr, objinstptr);
 
-extern float UDrawChar(u_char, short, short, int, int, float);
+void UDrawCharString(u_char *text, int start, int end, XfPoint *offset,
+     short styles, short ffont, int groupheight, int passcolor, float tmpscale);
 extern void UDrawString(labelptr, int, objinstptr);
 extern void UDrawStringNoX(labelptr, int, objinstptr);
 extern void CheckMarginStop(labelptr, objinstptr, Boolean);
@@ -1032,7 +1031,5 @@ extern void makecursors(void);
 /* from cairo.c */
 
 void xc_cairo_set_matrix(const Matrix *xcm);
-void xc_cairo_set_color(unsigned long coloridx);
-float xc_cairo_get_char_extents(const fontinfo *font, unsigned char c,
-      float *top, float *bottom);
-
+void xc_cairo_set_color(int coloridx);
+void xc_cairo_set_fontinfo(size_t fontidx);
