@@ -48,11 +48,11 @@ extern Tcl_Interp *xcinterp;
 
 extern Globaldata xobjs;
 extern XCWindowData *areawin;
-extern int	  *appcolors;
 extern xcWidget     menuwidgets[];
 extern xcWidget	  netbutton;
 extern char	  _STR[150];
 extern char	  _STR2[250];
+extern colorindex *colorlist;
 
 /*-------------------------------------------------------------------------*/
 
@@ -515,37 +515,37 @@ void setsymschem()
 
    if (topobject->schemtype == PRIMARY || topobject->schemtype == SECONDARY) {
       if (topobject->symschem == NULL) {
-         XtSetArg(aargs[0], XtNbackground, OFFBUTTONCOLOR);
-	 XtSetArg(aargs[1], XtNforeground, OFFBUTTONCOLOR);
+         XtSetArg(aargs[0], XtNbackground, colorlist[OFFBUTTONCOLOR].color.pixel);
+	 XtSetArg(aargs[1], XtNforeground, colorlist[OFFBUTTONCOLOR].color.pixel);
       }
       else {
-         XtSetArg(aargs[0], XtNbackground, BACKGROUND);
-	 XtSetArg(aargs[1], XtNforeground, FOREGROUND);
+         XtSetArg(aargs[0], XtNbackground, colorlist[BACKGROUND].color.pixel);
+	 XtSetArg(aargs[1], XtNforeground, colorlist[FOREGROUND].color.pixel);
       }
 
-      XtSetArg(bargs[1], XtNforeground, FOREGROUND);
-      XtSetArg(bargs[0], XtNbackground, SNAPCOLOR);
+      XtSetArg(bargs[1], XtNforeground, colorlist[FOREGROUND].color.pixel);
+      XtSetArg(bargs[0], XtNbackground, colorlist[SNAPCOLOR].color.pixel);
    }
    else {
       if (topobject->symschem != NULL) {
-         XtSetArg(bargs[0], XtNbackground, BACKGROUND);
-	 XtSetArg(bargs[1], XtNforeground, FOREGROUND);
+         XtSetArg(bargs[0], XtNbackground, colorlist[BACKGROUND].color.pixel);
+	 XtSetArg(bargs[1], XtNforeground, colorlist[FOREGROUND].color.pixel);
       }
       else {
-         XtSetArg(bargs[0], XtNbackground, OFFBUTTONCOLOR);
-	 XtSetArg(bargs[1], XtNforeground, OFFBUTTONCOLOR);
+         XtSetArg(bargs[0], XtNbackground, colorlist[OFFBUTTONCOLOR].color.pixel);
+	 XtSetArg(bargs[1], XtNforeground, colorlist[OFFBUTTONCOLOR].color.pixel);
       }
 
-      XtSetArg(aargs[1], XtNforeground, FOREGROUND);
+      XtSetArg(aargs[1], XtNforeground, colorlist[FOREGROUND].color.pixel);
       if (topobject->schemtype == FUNDAMENTAL)
-         XtSetArg(aargs[0], XtNbackground, AUXCOLOR);
+         XtSetArg(aargs[0], XtNbackground, colorlist[AUXCOLOR].color.pixel);
       else if (topobject->schemtype == TRIVIAL || topobject->symschem != NULL)
-         XtSetArg(aargs[0], XtNbackground, SNAPCOLOR);
+         XtSetArg(aargs[0], XtNbackground, colorlist[SNAPCOLOR].color.pixel);
       else {
-         XtSetArg(aargs[0], XtNbackground, OFFBUTTONCOLOR);
-	 XtSetArg(aargs[1], XtNforeground, OFFBUTTONCOLOR);
-         XtSetArg(bargs[0], XtNbackground, BBOXCOLOR);
-         XtSetArg(bargs[1], XtNforeground, FOREGROUND);
+         XtSetArg(aargs[0], XtNbackground, colorlist[OFFBUTTONCOLOR].color.pixel);
+	 XtSetArg(aargs[1], XtNforeground, colorlist[OFFBUTTONCOLOR].color.pixel);
+         XtSetArg(bargs[0], XtNbackground, colorlist[BBOXCOLOR].color.pixel);
+         XtSetArg(bargs[1], XtNforeground, colorlist[FOREGROUND].color.pixel);
       }
    }
 
@@ -862,7 +862,7 @@ void swapschem(int allow_create, int libnum, char *fullname)
 	 /* Push the current instance onto the push stack */
 	 /* Change the page without destroying the pushlist */
 
-	 push_stack(&areawin->stack, areawin->topinstance);
+	 push_stack(&areawin->stack, areawin->topinstance, NULL);
 	 stacktop = areawin->stack;
 	 areawin->stack = NULL;
          changepage(tpage);
@@ -893,7 +893,7 @@ void swapschem(int allow_create, int libnum, char *fullname)
 
 	 topobject->viewscale = areawin->vscale;
 	 topobject->pcorner = areawin->pcorner;
-	 push_stack(&areawin->stack, areawin->topinstance);
+	 push_stack(&areawin->stack, areawin->topinstance, NULL);
 	 areawin->topinstance = addtoinstlist(loclibnum, *newobject, FALSE);
 
 	 /* Generate the default bounding box for a size-zero object */
@@ -1014,7 +1014,7 @@ void swapschem(int allow_create, int libnum, char *fullname)
 
 	 topobject->viewscale = areawin->vscale;
 	 topobject->pcorner = areawin->pcorner;
-	 push_stack(&areawin->stack, areawin->topinstance);
+	 push_stack(&areawin->stack, areawin->topinstance, NULL);
 	 areawin->topinstance = syminst;
       }
    }
