@@ -42,11 +42,12 @@
 /*----------------------------------------------------------------------*/
 
 extern Display	*dpy;
-extern int *appcolors;
 extern Cursor	appcursors[NUM_CURSORS];
 extern XCWindowData *areawin;
 extern Globaldata xobjs;
 extern char _STR[150];
+extern int number_colors;
+extern colorindex *colorlist;
 
 #ifdef TCL_WRAPPER 
 extern Tcl_Interp *xcinterp;
@@ -110,15 +111,18 @@ void selectfilter(xcWidget w, pointertype value, caddr_t calldata)
 
 Boolean checkselect_draw(short value, Boolean draw_selected)
 {
-   short *check, savemode;
+   short *check;
  
    value &= areawin->filter;	/* apply the selection filter */
 
    if (areawin->selects == 0) {
-      savemode = eventmode;
-      if (!draw_selected) eventmode = PENDING_MODE;
-      select_element(value);
-      eventmode = savemode;
+      if (draw_selected)
+         select_element(value);
+      else {
+         Boolean save_redraw = areawin->redraw_needed;
+         select_element(value);
+         areawin->redraw_needed = save_redraw;
+      }
    }
    if (areawin->selects == 0) return False;
    for (check = areawin->selectlist; check < areawin->selectlist +
@@ -204,35 +208,15 @@ void geneasydraw(short instance, int mode, objectptr curobj, objinstptr curinst)
 
 void gendrawselected(short *newselect, objectptr curobj, objinstptr curinst)
 {
-   /* Don't draw selection color when selecting for edit */
-   if (eventmode == PENDING_MODE) return;
-
    if (*newselect >= curobj->parts) return;	// Safety check
 
-#ifndef HAVE_CAIRO
-   XcSetFunction(GXcopy);
-   XcSetForeground(BACKGROUND);
-   geneasydraw(*newselect, DOFORALL, curobj, curinst);
-   indicateparams(*(curobj->plist + *newselect));
-#endif /* !HAVE_CAIRO */
-
-#ifdef HAVE_CAIRO
    XcSetForeground(SELECTCOLOR);
-#else
-   XcSetFunction(GXxor);
-   XcSetForeground(SELECTCOLOR ^ BACKGROUND);
-#endif /* HAVE_CAIRO */
    geneasydraw(*newselect, DOFORALL, curobj, curinst);
 
-#ifdef HAVE_CAIRO
    SetForeground(dpy, areawin->gc, AUXCOLOR);
-#else
-   SetForeground(dpy, areawin->gc, AUXCOLOR ^ BACKGROUND);
-#endif /* HAVE_CAIRO */
    indicateparams(*(curobj->plist + *newselect));
 
    SetForeground(dpy, areawin->gc, areawin->gccolor);
-   SetFunction(dpy, areawin->gc, areawin->gctype);
 }
 
 /*---------------------------------------------------*/
@@ -1187,9 +1171,6 @@ void draw_normal_selected(objectptr thisobj, objinstptr thisinst)
 
    saveselects = areawin->selects;
 
-   /* reset the graphics state */
-   SetFunction(dpy, areawin->gc, GXcopy);
-
    areawin->selects = 0;
    drawarea(NULL, NULL, NULL);
    areawin->selects = saveselects;
@@ -1223,8 +1204,10 @@ static void freeselection(selection *rselect)
 
 void freeselects()
 {
-   if (areawin->selects > 0)
+   if (areawin->selects > 0) {
       free(areawin->selectlist);
+      areawin->redraw_needed =True;
+   }
    areawin->selects = 0;
    free_stack(&areawin->hierstack);
 }
@@ -1276,7 +1259,6 @@ void unselect_all()
    if (xobjs.suspend < 0)
       draw_normal_selected(topobject, areawin->topinstance);
    clearselects();
-
 }
 
 /*----------------------------------------------------------------------*/
@@ -1389,7 +1371,7 @@ void startselect()
    eventmode = SELAREA_MODE;
    areawin->origin.x = areawin->save.x;
    areawin->origin.y = areawin->save.y;
-   UDrawBox(areawin->origin, areawin->save);
+   selarea_mode_draw(xcDRAW_INIT, NULL);
 
 #ifdef TCL_WRAPPER
    Tk_CreateEventHandler(areawin->area, ButtonMotionMask |
@@ -1415,11 +1397,9 @@ void trackselarea()
    newpos = UGetCursorPos();
    if (newpos.x == areawin->save.x && newpos.y == areawin->save.y) return;
 
-   UDrawBox(areawin->origin, areawin->save);
-   UDrawBox(areawin->origin, newpos);
-
    areawin->save.x = newpos.x;
    areawin->save.y = newpos.y;
+   selarea_mode_draw(xcDRAW_EDIT, NULL);
 }
 
 /*----------------------*/
@@ -1433,11 +1413,9 @@ void trackrescale()
    newpos = UGetCursorPos();
    if (newpos.x == areawin->save.x && newpos.y == areawin->save.y) return;
 
-   UDrawRescaleBox(&areawin->save);
-   UDrawRescaleBox(&newpos);
-
    areawin->save.x = newpos.x;
    areawin->save.y = newpos.y;
+   rescale_mode_draw(xcDRAW_EDIT, NULL);
 }
 
 /*----------------------------------------------------------------------*/
@@ -1696,9 +1674,7 @@ short *recurse_select_element(short class, u_char mode) {
       }
       if (ilast >= 0) {
 	 newselect = rselect->selectlist + ilast;
-         SetFunction(dpy, areawin->gc, GXcopy);
-         XTopSetForeground(SELTOCOLOR(newselect));
-         geneasydraw(*newselect, DEFAULTCOLOR, topobject, areawin->topinstance);
+	 areawin->redraw_needed = True;
          areawin->selects--;
          for (k = jlast; k < areawin->selects; k++)
 	    *(areawin->selectlist + k) = *(areawin->selectlist + k + 1);
@@ -1735,7 +1711,7 @@ short *recurse_select_element(short class, u_char mode) {
       if (ilast >= 0) {
          newselect = allocselect();
          *newselect = *(rselect->selectlist + ilast);
-         gendrawselected(newselect, topobject, areawin->topinstance);
+	 areawin->redraw_needed = True;
          setoptionmenu();
          u2u_snap(&areawin->save);
 
