@@ -1026,7 +1026,9 @@ short undo_one_action()
 	       thisgraphic = (graphicptr)egen;
 	       escale->scale = thisgraphic->scale;
 	       thisgraphic->scale = fnum;
+#ifndef HAVE_CAIRO
 	       thisgraphic->valid = FALSE;
+#endif /* HAVE_CAIRO */
 	       break;
 	    case LABEL:
 	       thislabel = (labelptr)egen;
@@ -1450,7 +1452,9 @@ short redo_one_action()
 	       thisgraphic = (graphicptr)egen;
 	       escale->scale = thisgraphic->scale;
 	       thisgraphic->scale = fnum;
+#ifndef HAVE_CAIRO
 	       thisgraphic->valid = FALSE;
+#endif /* HAVE_CAIRO */
 	       break;
 	    case LABEL:
 	       thislabel = (labelptr)egen;
